import json
import logging
import time
from typing import Tuple, List, Any

import krakenex
import zmq
from requests import HTTPError
from schema import Schema

from krakenforwarder.util import FORMAT_INTERNAL_ZMQ_PUBLISH_TRADE, KEY_KRAKEN_ASSET_PAIR, KEY_KRAKEN_ERROR, \
    KEY_KRAKEN_LAST, KEY_KRAKEN_PAIR, KEY_KRAKEN_RESULT, KEY_KRAKEN_SINCE, KEY_KRAKEN_TRADES, KEY_PULL_PERIOD, \
    KEY_ZMQ_PUBLISH_PORT, VALUE_INTERNAL_OVER, VALUE_INTERNAL_TRADE

__all__ = ['KrakenForwarder']


class KrakenForwarder:
    def __init__(self, config: dict):
        Schema({
            KEY_PULL_PERIOD: int,
            KEY_KRAKEN_ASSET_PAIR: str,
            KEY_ZMQ_PUBLISH_PORT: int,
        }, ignore_extra_keys=True).validate(config)
        self.__pull_period = config[KEY_PULL_PERIOD]
        self.__kraken_asset_pair = config[KEY_KRAKEN_ASSET_PAIR]
        self.__zmq_publish_port = config[KEY_ZMQ_PUBLISH_PORT]
        self.__socket = None  # ZMQ not initiated -> must be initiated in subprocess
        self.__kraken = krakenex.API()
        self.__kraken_since = None
        self.__last_time_done = 0
        self.__logger = logging.getLogger()

        self.__logger.info(
            'core initiated'
            ' with {key_pull_period}={value_pull_period},'
            ' {key_asset_pair}={value_asset_pair}'
            ' and {zmq_publish}={value_zmq_publish}'.format(
                key_pull_period=KEY_PULL_PERIOD,
                value_pull_period=self.__pull_period,
                key_asset_pair=KEY_KRAKEN_ASSET_PAIR,
                value_asset_pair=self.__kraken_asset_pair,
                zmq_publish=KEY_ZMQ_PUBLISH_PORT,
                value_zmq_publish=self.__zmq_publish_port,
            )
        )

    def __zmq_init(self) -> None:  # ZMQ has to be called in subprocess
        context = zmq.Context()
        self.__socket = context.socket(zmq.PUB)
        self.__socket.bind('tcp://*:{port}'.format(port=self.__zmq_publish_port))

    def forward(self) -> None:  # this is the core loop
        self.__zmq_init()
        while True:
            current_time = int(time.time())  # current time in seconds

            # wait, eventually
            if current_time % self.__pull_period or current_time == self.__last_time_done:
                time.sleep(0.1)
                continue

            # get recent trades
            try:
                recent_trades, self.__kraken_since = self.__pull_recent_trades()
                self.__last_time_done = current_time
            except (ConnectionError, HTTPError) as error_msg:
                self.__logger.error(error_msg)
                continue

            self.__logger.info('Got {number_of_trades} trades'.format(number_of_trades=len(recent_trades)))
            if len(recent_trades) > 0:
                # broadcast
                for trade in recent_trades:
                    msg = FORMAT_INTERNAL_ZMQ_PUBLISH_TRADE.format(
                        label=VALUE_INTERNAL_TRADE,
                        pair=self.__kraken_asset_pair,
                        trade_info=trade
                    )
                    self.__socket.send_string(msg)
                self.__socket.send_string(VALUE_INTERNAL_OVER)

    def __pull_recent_trades(self) -> Tuple[List[List[Any]], int]:
        query_result = self.__kraken.query_public(
            KEY_KRAKEN_TRADES, {
                KEY_KRAKEN_PAIR: self.__kraken_asset_pair,
                KEY_KRAKEN_SINCE: self.__kraken_since
            })

        if len(query_result[KEY_KRAKEN_ERROR]) > 0:
            raise ConnectionError(json.dumps(query_result[KEY_KRAKEN_ERROR]))

        return query_result[KEY_KRAKEN_RESULT][self.__kraken_asset_pair], \
               int(query_result[KEY_KRAKEN_RESULT][KEY_KRAKEN_LAST])



import zmq

from krakenforwarder.util import KEY_ZMQ_PUBLISH_PORT, KEY_ZMQ_HOSTNAME

__all__ = ['listen']


def listen(config: dict):
    context = zmq.Context()
    socket = context.socket(zmq.SUB)
    socket.connect('tcp://{zmq_host}:{zmq_port}'.format(
        zmq_host=config[KEY_ZMQ_HOSTNAME],
        zmq_port=config[KEY_ZMQ_PUBLISH_PORT])
    )
    socket.setsockopt_string(zmq.SUBSCRIBE, '')
    while True:
        msg = socket.recv_string()
        yield msg



from typing import List


def merge_dicts(dics: List[dict]) -> dict:
    """
    Merges a list of dict objects to one dict object
    """
    merged = dict()
    for dic in dics:
        merged.update(dic)
    return merged


FORMAT_INTERNAL_ZMQ_PUBLISH_TRADE = '{label}::{pair}::{trade_info}'
KEY_KRAKEN_ASSET_PAIR = 'Asset Pair'
KEY_KRAKEN_ERROR = 'error'
KEY_KRAKEN_LAST = 'last'
KEY_KRAKEN_PAIR = 'pair'
KEY_KRAKEN_RESULT = 'result'
KEY_KRAKEN_SINCE = 'since'
KEY_KRAKEN_TRADES = 'Trades'
KEY_PULL_PERIOD = 'Pull Period'
KEY_ZMQ_PUBLISH_PORT = 'Publish Port'
KEY_ZMQ_HOSTNAME = 'Hostname'
VALUE_INTERNAL_OVER = 'OVER'
VALUE_INTERNAL_TRADE = 'TRADE'



from setuptools import setup, find_packages

import krakenforwarder

setup(
    name='krakenforwarder',
    version=krakenforwarder.__version__,
    packages=find_packages(exclude=('tests',)),
    include_package_data=True,
    url='https://github.com/supercuiller/krakenforwarder',
    license='MIT',
    author='Thibaut Castaings',
    author_email='castaings.t@protonmail.com',
    description='Forwards trades from Kraken.com to TCP port',
    long_description=open('README.md').read(),
    install_requires=[
              'pyzmq',
              'schema',
              'pyyaml',
              'krakenex'
          ],
)
